/*
 * scope3d main code.
 *
 * This is hacky as first implementation didn't look so great.  That first
 * impl. created a display list (wasting precious RAM) with a first pass, then
 * displayed it.  I then added DRAW_WHILE_THINKING; this sends output to the DAC
 * as new points are calculated, spreading out the 'thinking time'.
 *
 * (c) 2012 Matt Evans
 * Licenced under GPLv2
 */

#include <inttypes.h>
#include "hw.h"
/* Define CPU frequency before including delay.h! */
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h> // CPU freq. defined!
#include <stdlib.h>

#include "serio.h"
#include "itrig.h"
#include "dac.h"
#include "points.h"

/* Try to break down long lines into 10-unitish length
 * pieces.  Note I don't do a square root here so... it's
 * all a bit inaccurate and varied. ;-)
 */
#define SUBDIV		15

#define MAX_LINES	200

#define SCREEN_X	256
#define SCREEN_Y	256

#define DRAW_WHILE_THINKING

/*#define LINES_DEBUG*/

#define EB_OFF()	do { PORTC |= B(PC5); } while(0)
#define EB_ON()		do { PORTC &= ~B(PC5); } while(0)

#define MAX(x, y)	((x) > (y) ? (x) : (y))
#define SAT_INT16(x)	(((x) > 0x7000L) ? 0x7000 : (((x) < -0x7000L) ? -0x7000 : (x)))


/*
 * Internal data representation:
 *
 * The main loop asks to draw_line(x,y) with int16_t resolution.  To
 * save memory, these lines are put in the display list with 8-bit
 * screen coordinates.  (Unsigned; (0,0) to (255,255) bottom left to
 * top right.)  The display routine subdivides long lines with higher
 * 10-bit resolution (to match the 10-bit DAC).  Subdivision is done
 * so the electron beam doesn't move TOO fast; it's all very rough
 * guesswork anyway.
 *
 * The display list is of the format:
 * [8 bits]	flags:
 * 	0	End
 *	1	Line continues on from last point
 *	2	Blank beam, move to X,Y
 * [8 bits]	X
 * [8 bits]	Y
 *
 * Thus, in, say, 768 bytes we can store 256 line segments.
 *
 * All this means that lines must come pre-clipped, endpoints off-screen are
 * not supported.  So calculate magic 3D stuff in 16bit fixed-point then
 * clip whilst rounding down to 8-bit screen res.
 */

#ifndef DRAW_WHILE_THINKING
point_t disp_list[MAX_LINES];
#endif
uint8_t cur_line = 0;

/* Number of internal in-flash objects: */
#define NOBJS	2

PROGMEM point3d def_none[] = {
	{ TOK_SPECIAL, TOK_LINES, 1 },
        { -100, -100, -100 }, {  100,  100,  100 },
        { TOK_SPECIAL, TOK_END, 0 }
};

PROGMEM point3d cube[] = {
#include "cube_byhand.inc"
};

extern PROGMEM point3d tie[];

#ifdef DRAW_WHILE_THINKING
#define DL_DATA_SIZE	300	/* Big model! */
uint8_t dl_data[DL_DATA_SIZE];
uint16_t offs = 0;
#endif

static point_t last_point = { .flags = 0, .x = 0, .y = 0 };

/* Primitives for DAC line drawing-- move and output a line.
 */
static void line_moveto(uint8_t x, uint8_t y)
{
	EB_OFF();
	/* New move.  Just move there immediately. */
	dac_write(x << 2, y << 2);
	/* Should scale this delay by line length,
	 * output RC filter doesn't let the output change
	 * very quickly sometimes...
	 *
	 * If we can't blank... this doesn't
	 * matter. */
	_delay_us(20);
	last_point.x = x;
	last_point.y = y;
}

static void line_drawto(uint8_t new_x, uint8_t new_y)
{
	/* New line.  How long is it?  Output steps.. */
	uint16_t max_dim = MAX(abs(new_x - last_point.x),
			       abs(new_y - last_point.y));
#ifdef LINES_DEBUG
	PS("Draw to ");
	phex8(new_x);
	putchr(',');
	phex8(new_y);
	pnewl();
#endif
	if (max_dim < SUBDIV) {
		/* Short line, just move right away */
		dac_write(new_x << 2, new_y << 2);
		EB_ON();
	} else {
		uint8_t steps = idiv(max_dim, SUBDIV);
		uint8_t i;
		/* x,y in 10:5 fixed point (+sign at top!) */
		int16_t x = last_point.x << (2+5);
		int16_t y = last_point.y << (2+5);
		/* The sign of this division is important! */
		int16_t dx = idiv((((int16_t)new_x << (2+5))
				   - x), (steps+1));
		int16_t dy = idiv((((int16_t)new_y << (2+5))
				   - y), (steps+1));
#ifdef LINES_DEBUG
		PS(" Steps ");
		phex8(steps);
		PS(", x,y = ");
		phex16(x);
		putchr(',');
		phex16(y);
		PS(" dx,dy = ");
		phex16(dx);
		putchr(',');
		phex16(dy);
		pnewl();
#endif
		for (i = 0; i < (steps+1); i++) {
			x += dx;
			y += dy;
			dac_write(x >> 5, y >> 5);
			EB_ON();
#ifdef LINES_DEBUG
			PS("  -subdiv x,y = ");
			phex16(x);
			putchr(',');
			phex16(y);
			PS(" scaled x,y = ");
			phex8(x >> (2+5));
			putchr(',');
			phex8(y >> (2+5));
			pnewl();
#endif
		}
		/* And hopefully at the end, x,y = new_point! */
		/* *Whistle*... hack */
		if ((x >> (2+5) != new_x) ||
		    (y >> (2+5) != new_y)) {
			dac_write(new_x << 2,
				  new_y << 2);
			EB_ON();
#ifdef LINES_DEBUG
			PS(" click over to ");
			phex8(new_x);
			putchr(',');
			phex8(new_y);
			pnewl();
#endif
		}
	}
	last_point.x = new_x;
	last_point.y = new_y;
}


/* Cohen-Sutherland line clipping */
#define LEFT		1
#define RIGHT		2
#define BOTTOM		4
#define TOP		8
static uint8_t outcode(int16_t x, int16_t y)
{
	uint8_t code = 0;
	if (x < 0)
		code |= LEFT;
	else if (x > (SCREEN_X-1))
		code |= RIGHT;
	if (y < 0)
		code |= BOTTOM;
	else if (y > (SCREEN_Y-1))
		code |= TOP;
	return code;
}

/* Add a line to the display list.  Careful of ordering!  This function will
 * try to optimise follow-on lines, e.g. dlist_line(0,0,  100,100) then
 * dlist_line(100,100,  150,100) will add a move plus two line_to entries.
 * Calling them the other way round will add a move, a line_to, another move, another
 * line_to.
 *
 * Output coordinates are unsigned 8-bit quantities.  Input coordinates are signed
 * 16-bit -- screen is still (0,0) -> (255,255) in this space but we can now
 * do clipping.
 */
static void dlist_line(int16_t x1, int16_t y1, int16_t x2, int16_t y2)
{
	static int16_t last_x = -32768;
	static int16_t last_y = -32768;
	uint8_t c1 = outcode(x1, y1);
	uint8_t c2 = outcode(x2, y2);
	uint8_t accept = 0;

	if (cur_line >= MAX_LINES) {
		PS("Line overflow!\r\n");
		return;
	}

#ifdef LINES_DEBUG
	PS("dlist_line ");
#endif
	while (1) {
#ifdef LINES_DEBUG
		PS("x1,y1 x2,y2 = ");
		phex16(x1);
		putchr(',');
		phex16(y1);
		putchr(' ');
		phex16(x2);
		putchr(',');
		phex16(y2);
		PS(" c1,c2 = ");
		phex8(c1);
		putchr(',');
		phex8(c2);
		pnewl();
#endif
		if ((c1 | c2) == 0) {
			/* Both ends inside screen -- carry on */
			accept = 1;
			break;
		} else if ((c1 & c2) != 0) {
			/* Line off-screen, so ignore. */
			break;
		} else {
			/* Clip */
			int16_t x = 0, y = 0;

#ifdef LINES_DEBUG
			PS(" clip! ");
#endif

			/* Cohen-Sutherland algorithm.  This may loop up to
			 * twice; a line extending from, e.g. bottom of screen,
			 * across screen, to top will have 2 intersection
			 * points.
			 */
			uint8_t c = c1 ? c1 : c2;

			if (c & TOP) {
				/* point is above */
				x = x1 + idiv(((x2 - x1) * (SCREEN_Y-1 - y1)),
					      (y2 - y1));
				y = SCREEN_Y-1;
			} else if (c & BOTTOM) {
				/* point is below */
				x = x1 + idiv(((x2 - x1) * (0 - y1)),
					      (y2 - y1));
				y = 0;
			} else if (c & RIGHT) {
				/* point is to the right */
				y = y1 + idiv(((y2 - y1) * ((SCREEN_X-1) - x1)),
					      (x2 - x1));
				x = SCREEN_X-1;
			} else if (c & LEFT) {
				/* point is to the left */
				y = y1 + idiv(((y2 - y1) * (0 - x1)),
					      (x2 - x1));
				x = 0;
			}
			/* Now we move outside point to intersection point to
			 * clip and get ready for next pass.
			 */
			if (c == c1) {
				x1 = x;
				y1 = y;
				c1 = outcode(x1, y1);
			} else {
				x2 = x;
				y2 = y;
				c2 = outcode(x2, y2);
			}
		}
	}

	if (accept) {
		if (x1 == last_x && y1 == last_y) {
#ifdef LINES_DEBUG
			PS(" run-on to ");
			phex16(x2);
			putchr(',');
			phex16(y2);
#endif

#ifdef DRAW_WHILE_THINKING
			line_drawto(x2, y2);
#else
			disp_list[cur_line].flags = FLAG_LINE_TO;
			disp_list[cur_line].x = x2;
			disp_list[cur_line].y = y2;
#endif
			last_x = x2;
			last_y = y2;
			cur_line++;
		} else {
			/* Could test x2/y2 as well... */
			if (cur_line < (MAX_LINES-1)) {
#ifdef LINES_DEBUG
				PS(" move to ");
				phex16(x1);
				putchr(',');
				phex16(y1);
				PS(" draw to ");
				phex16(x2);
				putchr(',');
				phex16(y2);
#endif
#ifdef DRAW_WHILE_THINKING
				line_moveto(x1, y1);
				line_drawto(x2, y2);
#else
				disp_list[cur_line].flags = FLAG_MOVE_TO;
				disp_list[cur_line].x = x1;
				disp_list[cur_line].y = y1;
				cur_line++;
				disp_list[cur_line].flags = FLAG_LINE_TO;
				disp_list[cur_line].x = x2;
				disp_list[cur_line].y = y2;
#endif
				last_x = x2;
				last_y = y2;
				cur_line++;
			} else {
				PS("Line ovf on move-line\r\n");
			}
		}
#ifdef LINES_DEBUG
		pnewl();
#endif
	}
}

static void dlist_done(void)
{
#ifndef DRAW_WHILE_THINKING
	if (cur_line >= MAX_LINES) {
		PS("Done overflow!\r\n");
		return;
	}
	disp_list[cur_line].flags = FLAG_END;
	cur_line++;
#endif
}

#define S_NONE			0
#define S_LS_START		1
#define S_LS_MID		2
#define S_L_START		3
#define S_L_END			4

static void generate_dlist(point3d *obj, uint8_t in_flash,
			   unsigned int angx, unsigned int angy,
			   unsigned int angz)
{
	uint8_t i = 0;
	uint8_t done = 0;
	uint8_t state = S_NONE;
	uint8_t count = 0;

	int16_t slast_x = 0, slast_y = 0;

	/* A 254*254*254 cube corner is this far from the centre (0,0,0),
	 * so this makes z always positive...*/
#if 1
	/* This is actually more correct. */
	int16_t zoffs = 222;
	int16_t d = 700;
#else
	int16_t zoffs = 10;
	int16_t d = 222;
#endif

	cur_line = 0;

	do {
		int16_t tx, ty, tz, tn;
		int16_t sx, sy;
		int8_t ox, oy, oz;

		do {
			/* Read point in list */
			if (in_flash) {
				ox = (int8_t)pgm_read_byte(&obj[i].x);
				oy = (int8_t)pgm_read_byte(&obj[i].y);
				oz = (int8_t)pgm_read_byte(&obj[i].z);
			} else {
				/* Safety pressure valve, don't loop
				 * doing nothing useful if we're given
				 * bad data.
				 */
				if ((i*3) > DL_DATA_SIZE)
					done = 1;
				ox = obj[i].x;
				oy = obj[i].y;
				oz = obj[i].z;
			}
#ifndef DRAW_WHILE_THINKING
			/* This gives the EB something to do if we can't
			 * blank it... (e.g. the OS1000 scope :( )
			 */
			dac_write(ox ^ oy<<2 ^ oz, (i < 128) ? 0 : 255<<2);
#endif

#ifdef LINES_DEBUG
			PS("read ");
			phex8(i);
			PS(", xyz = ");
			phex32(ox);
			putchr(',');
			phex32(oy);
			putchr(',');
			phex32(oz);
			pnewl();
#endif
			i++;

			/* TODO: Add 2D tokens to this escape, so
			 * can download/display untransformed 2D
			 */

			if ((uint8_t)ox == TOK_SPECIAL) {
				/* Escape coordinate */
				switch ((uint8_t)oy) {
				case TOK_END:
#ifdef LINES_DEBUG
					PS("END\r\n");
#endif
					done = 1;
					break;
				case TOK_LINE_STRIP:
#ifdef LINES_DEBUG
					PS("LS start\r\n");
#endif
					state = S_LS_START;
					count = oz;
					break;
				case TOK_LINES:
#ifdef LINES_DEBUG
					PS("Lines start\r\n");
#endif
					state = S_L_START;
					count = oz;
					break;
				default:
					PS("Unknown special token ");
					phex8(oy);
					pnewl();
					done = 1;
				}
			}
		} while (!done && (uint8_t)ox == TOK_SPECIAL);

		if (!done) {
			tx = ox;
			ty = oy;
			tz = oz;
			// Z
			if ((angz & (TWOPI-1)) != 0) {
				tn = ((tx * (int32_t)isin(angz)) +
				      (ty * (int32_t)icos(angz)))>>15;
				tx = ((tx * (int32_t)icos(angz)) -
				      (ty * (int32_t)isin(angz)))>>15;
				ty = tn;
			}
			// X
			if ((angx & (TWOPI-1)) != 0) {
				tn = ((tz * (int32_t)isin(angx)) +
				      (ty * (int32_t)icos(angx)))>>15;
				tz = ((tz * (int32_t)icos(angx)) -
				      (ty * (int32_t)isin(angx)))>>15;
				ty = tn;
			}
			// Y
			if ((angy & (TWOPI-1)) != 0) {
				tn = ((tx * (int32_t)isin(angy)) +
				      (tz * (int32_t)icos(angy)))>>15;
				tx = ((tx * (int32_t)icos(angy)) -
				      (tz * (int32_t)isin(angy)))>>15;
				tz = tn;
			}
			tz += zoffs;

			if (tz < 0)
				tz = 0;

			sx = idiv(d*(int32_t)tx, d+tz);
			sy = idiv(d*(int32_t)ty, d+tz);

#ifdef LINES_DEBUG
			PS("Point ");
			phex32(tx);
			putchr(',');
			phex32(ty);
			putchr(',');
			phex32(tz);
			PS(" transformed to ");
			phex32(sx);
			putchr(',');
			phex32(sy);
			pnewl();
#endif

			switch (state) {
			case S_LS_START:
				/* This macro chops 32bit to 16, saturating to
				 * MAX/MIN_SHORT if not in 16b range
				 */
				slast_x = SAT_INT16(128L+sx);
				slast_y = SAT_INT16(128L+sy);
				state = S_LS_MID;
				count--;
				break;
			case S_L_START:
				slast_x = SAT_INT16(128L+sx);
				slast_y = SAT_INT16(128L+sy);
				state = S_L_END;
				break;
			case S_LS_MID:
			{
				int16_t screenx = SAT_INT16(128L+sx);
				int16_t screeny = SAT_INT16(128L+sy);
				dlist_line(slast_x, slast_y, screenx, screeny);
				slast_x = screenx;
				slast_y = screeny;
				if (--count == 0)
					state = S_NONE;
			} break;
			case S_L_END:
			{
				int16_t screenx = SAT_INT16(128L+sx);
				int16_t screeny = SAT_INT16(128L+sy);
				dlist_line(slast_x, slast_y, screenx, screeny);
				slast_x = screenx;
				slast_y = screeny;
				if (count-- > 0)
					state = S_L_START;
				else
					state = S_NONE;
			} break;
			}
		}
	} while (!done);

	dlist_done();
}

#ifndef DRAW_WHILE_THINKING
static void draw_dlist(void)
{
	uint8_t n;

	EB_OFF();
	for (n = 0; n < MAX_LINES; n++) {
		point_t new_point = { .x = disp_list[n].x,
				      .y = disp_list[n].y };

		if (disp_list[n].flags == FLAG_END)
			break;
		else if (disp_list[n].flags == FLAG_MOVE_TO) {
#ifdef LINES_DEBUG
			PS("Move to ");
			phex8(new_point.x);
			putchr(',');
			phex8(new_point.y);
			pnewl();
#endif
			/* New move.  Just move there immediately. */
			line_moveto(new_point.x, new_point.y);
		} else if (disp_list[n].flags == FLAG_LINE_TO) {
			line_drawto(new_point.x, new_point.y);
		} else {
			PS("*** WTF? Bad flags ");
			phex8(disp_list[n].flags);
			PS(" at line offset ");
			phex8(n);
			pnewl();
		}
		last_point.x = new_point.x;
		last_point.y = new_point.y;
	}
	EB_OFF();
}
#endif

////////////////////////////////////////////////////////////////////////////////

void delay_cs(uint8_t s)
{
	for (; s > 0; s--) {
		_delay_ms(10);
	}
}

static void io_init(void)
{
	/* LED output */
	DDRB |= LEDG;
	/* Blanking */
	DDRC |= B(PC5);
}

////////////////////////////////////////////////////////////////////////////////
uint8_t	     	model = 0;
point3d 	*modelptr;
uint8_t 	model_in_flash = 1;

void set_model(uint8_t m)
{
	model = m;
	switch(model) {
	case 0:
		/* Tie test */
		modelptr = (point3d *)tie;
		model_in_flash = 1;
		break;
	case 1:
		modelptr = (point3d *)cube;
		model_in_flash = 1;
		break;
	case 255:
		modelptr = (point3d *)dl_data;
		model_in_flash = 0;
		break;
	default:
		modelptr = (point3d *)def_none;
		model_in_flash = 1;
	}
}

int main(void)
{
	uint16_t i = 0;
	unsigned int angx = 0;
	unsigned int angy = 0;
	unsigned int angz = 0;
	int8_t 	     d_angx = 7;
	int8_t 	     d_angy = 6;
	int8_t 	     d_angz = 0;
	char	     cmd = 0;
	uint8_t	     argc = 0;
	uint8_t	     argtmp = 0;

	io_init();
	serio_init();
	dac_init();

	sei();

	LEDG_ON;

	delay_cs(10);

	PS("\r\n\r\n----\r\nLovely vector generator V1.0\r\nY");

	set_model(0);

	do {
		uint8_t num;

		generate_dlist(modelptr, model_in_flash, angx, angy, angz);
#ifndef DRAW_WHILE_THINKING
		draw_dlist();
#endif

		/* Serial comms.:
		 * If we do, say, a minimum of 10Hz round this main loop,
		 * at 9600bps we'll get about 96 chars max to be parsed.
		 *
		 * It may be worth setting the model to cube first to get
		 * the FPS up so we poll serial more often & is more reliable
		 * for a download...!</hack>
		 * I'm trying to be stingy on RX buffer size. :P
		 *
		 * Support the following functions over serial:
		 * - Resync/idle
		 * - Download object data
		 * - Display object N (where 255 = downloaded object)
		 * - Set rotation x,y,z
		 * - Animate rotation x,y,z
		 *
		 * Serial format is ascii, [nn] = 8 bit ascii hex:
		 * ' '	 	- Resync, echoes 'Y' for success
		 * O[nn]    	- Display object hex [nn],
		 *		  echoes 'Y' for success, 'N' fail
		 * N		- Read num objects, returns hex [nn]'Y'
		 * R[xx][yy][zz]
		 *		- Set rotation x,y,z in units of 360/256 deg.
		 *		  Returns 'Y'
		 * A[xx][yy][zz]
		 *		- Set rotation animation, speed in each axis.
		 *		  Hint, rotation of 0 avoids the calculation
		 * 		  and will update quicker.
		 *		  Returns 'Y'
		 * F		- Read free object RAM (bytes)
		 *		  Returns [ll][hh]'Y' size available
		 * D[...]
		 *		- Download object data.  See how many bytes
		 * 		  you can write using 'F' first!
		 *		  Object data is written (in tuples of x,y,z
		 *		  coordinates, see comments elsewhere, and
		 *  		  see TOK_SPECIAL et al) after the 'D' but
		 *		  'END OF DATA' is signalled with a non-hex
		 *		  character, e.g. ' ' upon which case 'Y' is
		 *		  returned.
		 *		  Also, this is useful in the case that
		 *		  the link get out of sync; sending ' ' bytes
		 * 		  until you get a 'Y' response will break
		 *		  out of any download and return to idle.
		 */
		LEDG_ON;
		num = rx_data_ready();
		if (num != 0) {
			LEDG_OFF;
			if (cmd == 0) {
				cmd = getchr();
				argc = 0;
				argtmp = 0;
			}
			/* Any more? */
			num = rx_data_ready();
			switch (cmd) {
			case ' ':
				putchr('Y');
				cmd = 0;
				break;
			case 'O':
				if (num >= 1) {
					uint8_t h = gethex4();
					if (h != 0xff) {
						argtmp <<= 4;
						argtmp |= h;
						argc++;
						if (argc == 2) {
							set_model(argtmp);
							putchr('Y');
							/* Set angle, animation
							 * to 0?
							 */
							cmd = 0;
						}
					} else {
						putchr('N');
						cmd = 0;
					}
				}
				break;
			case 'N':
				phex8(NOBJS);
				putchr('Y');
				cmd = 0;
				break;
			case 'R':
				if (num >= 1) {
					uint8_t h = gethex4();
					if (h <= 0xf) {
						argtmp <<= 4;
						argtmp |= h;
						argc++;
						if (argc == 2) {
							angx = argtmp;
						} else if (argc == 4) {
							angy = argtmp;
						} else if (argc == 6) {
							angz = argtmp;
							putchr('Y');
							cmd = 0;
						}
					} else {
						putchr('N');
						cmd = 0;
					}
				}
				break;
			case 'A':
				if (num >= 1) {
					uint8_t h = gethex4();
					if (h <= 0xf) {
						argtmp <<= 4;
						argtmp |= h;
						argc++;
						if (argc == 2) {
							d_angx = argtmp;
						} else if (argc == 4) {
							d_angy = argtmp;
						} else if (argc == 6) {
							d_angz = argtmp;
							putchr('Y');
							cmd = 0;
						}
					} else {
						putchr('N');
						cmd = 0;
					}
				}
				break;
			case 'F':
				phex16(DL_DATA_SIZE);
				putchr('Y');
				cmd = 0;
				break;
			case 'D':
				if (num >= 1) {
					uint8_t h = gethex4();
					if (h <= 0xf) {
						argtmp <<= 4;
						argtmp |= h;
						argc++;
						if (argc == 2) {
							dl_data[offs++] = argtmp;
							argc = 0;
						}
					} else {
						/* Done */
						if (offs > 0)
							putchr('Y');
						else
							putchr('N');
						offs = 0;
						cmd = 0;
					}
				}
				break;
			default:
				putchr('N');
				cmd = 0;
			}
		}
		angx += d_angx;
		angy += d_angy;
		angz += d_angz;

		i++;
	}
#ifdef LINES_DEBUG
	while (++i < 10);
#else
	while (1);
#endif
	PS("End\r\n");
	return 0;
}
