/*
 * IRQ-driven serial RX/TX routines
 *
 * (c) 2012 Matt Evans
 * Licenced under GPLv2
 */

#include "hw.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "serio.h"

#define RX_SIZE 128
#define TX_SIZE 16

char tx_buffer[TX_SIZE];
char rx_buffer[RX_SIZE];

/* Circular buffer pointers.  WR points to the empty location next to be
 * written; rd points to the next thing to be read.
 * Empty if rx=wr. Full if wr=rx-1.
 */
volatile uint8_t rx_wr = 0;
volatile uint8_t rx_rd = 0;
volatile uint8_t tx_wr = 0;
volatile uint8_t tx_rd = 0;

uint16_t rx_dropped = 0;

extern void delay_cs(uint8_t n);

ISR(USART_RX_vect)
{
	char c = UDR0;

	if (((rx_wr+1) & (RX_SIZE-1)) == (rx_rd & (RX_SIZE-1))) {
		/* Arse! RX buffer is full. */
		rx_dropped++;
	} else {
		rx_buffer[rx_wr] = c;
		rx_wr = (rx_wr + 1) & (RX_SIZE-1);
	}
}

ISR(USART_UDRE_vect)
{
	/* Anything to send? */
	if (tx_wr != tx_rd) {
		UDR0 = tx_buffer[tx_rd];
		tx_rd = (tx_rd + 1) & (TX_SIZE-1);
	} else {
		/* Nope, all done.  Disable this IRQ so we don't loop. */
		UCSR0B &= ~B(UDRIE0);
	}
}

void putchr(char c)
{
	/* If output buffer full, stall. */
	while (((tx_wr+1) & (TX_SIZE-1)) == (tx_rd & (TX_SIZE-1)))
	{ }

	tx_buffer[tx_wr] = c;
	tx_wr = (tx_wr + 1) & (TX_SIZE-1);

	/* Kick-- if the IRQ is already enabled, fine, the IRQ handler
	 * will fire and send the next byte.
	 * If it's disabled (the handler's found there's nothing to do),
	 * this will re-enable and re-assess.
	 */
	UCSR0B |= B(UDRIE0);
}

char getchr(void)
{
	char c;
	/* Wait for a char to arrive if none. */
	while (!rx_data_ready())
	{ }

	c = rx_buffer[rx_rd];
	rx_rd = (rx_rd + 1) & (RX_SIZE-1);
	return c;
}

/* Return the number of bytes waiting. */
uint8_t rx_data_ready(void)
{
	return (rx_wr - rx_rd) & (RX_SIZE-1);
}

void serio_init(void)
{
	UCSR0A = 0;
	UCSR0B = (1<<TXEN0) | (1<<RXEN0) | B(RXCIE0);
	UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);
	UBRR0 = UBRR;
}

void pnewl(void)
{
	putchr('\r');
	putchr('\n');
}

/* printstr_p is from the AVR-libc examples */
void printstr_p(const char *s)
{
  char c;
  
  for (c = pgm_read_byte(s); c; ++s, c = pgm_read_byte(s))
  {
    if (c == '\n')
    {
      putchr('\r');
    }
    putchr(c);
  }
}

uint8_t gethex4(void)
{
	char c;

	c = getchr();
	if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	else if (c >= 'A' && c <= 'F')
		return c - 'A' + 10;
	else if (c >= '0' && c <= '9')
		return c - '0';
	else
		return 0xff;
}

void phex8(uint8_t n)
{
  uint8_t hi = (n >> 4);
  putchr(hi + ( (hi > 9) ? ('a' - 10) : '0' ));
  n = n & 0xf;
  putchr(n + ( (n > 9) ? ('a' - 10) : '0' ));
}

void phex16(uint16_t n)
{
  phex8(n >> 8);
  phex8(n & 0xff);
}

void phex32(uint32_t n)
{
  phex16(n >> 16);
  phex16(n & 0xffff);
}
